/* stablesort.c -- an adaptive, stable, natural mergesort

   http://py-stablesort.sourceforge.net/

   Wrapper code Copyright (C) 2002 Markus F.X.J. Oberhumer.

   Actual sort implementation Copyright (C) 2002 the Python Software
   Foundation. Written by Tim Peters.
 */


#define MODULE_VERSION  "1.1.2"

#include <Python.h>
#undef NDEBUG
#define NDEBUG
#include <assert.h>

#if !defined(PY_VERSION_HEX) || (PY_VERSION_HEX < 0x010502f0)
#  error "Need Python version 1.5.2 or better"
#endif

#undef UNUSED
#define UNUSED(var)     ((void)&var)


/***********************************************************************
// compat stuff
************************************************************************/

#if (PY_VERSION_HEX < 0x02020000)
#  define PyObject_Call(a,b,c)  PyEval_CallObjectWithKeywords(a,b,c)
#endif

#if (PY_VERSION_HEX < 0x02010000)
static int Compare_LT(PyObject *a, PyObject *b)
{
    int x;
    x = PyObject_Compare(a, b);
    if (PyErr_Occurred())
        return -1;
    return x < 0;
}
#endif


/***********************************************************************
// wrapper
************************************************************************/

static PyTypeObject *stablesort_immutable_list_type = NULL;

#define STABLESORT
#include "listobject.c"
#undef STABLESORT
#undef list


static PyMethodDef *list_sort_md = NULL;
static PyMethodDef saved_list_sort;


#define SORT_DOC \
"sort(list, [cmpfunc]) -- stable sort *IN PLACE*; cmpfunc(x, y) -> -1, 0, 1"
#define INSTALL_DOC \
"install() -- install the stablesort implementation as global list.sort() method"
#define UNINSTALL_DOC \
"uninstall() -- restore the original list.sort() implementation"

#define SORT_METH_DOC \
"L.sort([cmpfunc]) -- stable sort *IN PLACE*; if given, cmpfunc(x, y) -> -1, 0, 1"


/***********************************************************************
// stablesort.sort(list, cmpfunc=None)  [function]
************************************************************************/

static PyObject *
do_sort(PyObject *dummy, PyObject *args)
{
    PyListObject *list = NULL;
    PyObject *cmpfunc = NULL;

    UNUSED(dummy);
    if (!PyArg_ParseTuple(args, "O!|O:sort", &PyList_Type, &list, &cmpfunc))
        return NULL;
#if 1
    /* note: Python's list.sort() does _NOT_ accept None, but we do... */
    if (cmpfunc == Py_None)
        cmpfunc = NULL;
#endif
    return listsort(list, cmpfunc);
}


/***********************************************************************
// new list.sort()         [method]
// stablesort.install()    [function]
// stablesort.uninstall()  [function]
************************************************************************/

static PyObject *
do_new_list_sort(PyListObject *self, PyObject *args)
{
    PyObject *cmpfunc = NULL;

    if (!PyArg_ParseTuple(args, "|O:sort", &cmpfunc))
        return NULL;
#if 0
    /* note: Python's list.sort() does _NOT_ accept None, but we do... */
    if (cmpfunc == Py_None)
        cmpfunc = NULL;
#endif
    return listsort(self, cmpfunc);
}


static PyObject *
do_install(PyObject *dummy, PyObject *args)
{
    UNUSED(dummy);
    if (!PyArg_ParseTuple(args, ":install"))
        return NULL;
    if (list_sort_md->ml_meth != (PyCFunction)do_new_list_sort)
    {
        memcpy(&saved_list_sort, list_sort_md, sizeof(PyMethodDef));
        list_sort_md->ml_meth = (PyCFunction)do_new_list_sort;
        list_sort_md->ml_flags = METH_VARARGS;
        list_sort_md->ml_doc = SORT_METH_DOC;
    }
    Py_INCREF(Py_None);
    return Py_None;
}


static PyObject *
do_uninstall(PyObject *dummy, PyObject *args)
{
    UNUSED(dummy);
    if (!PyArg_ParseTuple(args, ":uninstall"))
        return NULL;
    if (list_sort_md->ml_meth == (PyCFunction)do_new_list_sort)
        memcpy(list_sort_md, &saved_list_sort, sizeof(PyMethodDef));
    Py_INCREF(Py_None);
    return Py_None;
}


/***********************************************************************
// dirty hack to get access to Python's internal list.sort() method
// and immutable_list_type
************************************************************************/

static PyObject *dummy_cmpfunc(PyObject *list, PyObject *args)
{
    UNUSED(args);
    if (list && list->ob_type && list->ob_type != &PyList_Type)
        stablesort_immutable_list_type = list->ob_type;
    return PyInt_FromLong(0);
}


static int init_internals(void)
{
    int result = -1;        /* guilty until proved innocent */
    PyObject *list = NULL;
    PyObject *cmpfunc = NULL;
    PyObject *sort_meth = NULL;
    PyObject *sort_args = NULL;
    PyObject *sort_res = NULL;
    PyMethodDef dummy_cmpfunc_md = {
        /* The name of this method does not matter, so we call it "sort"
         * in order to avoid allocating a new string object. */
        "sort", (PyCFunction)dummy_cmpfunc, METH_VARARGS, NULL
    };

    /* Create a list with 2 entries. */
    list = PyList_New(2);
    if (list == NULL)
        goto error;
    /* Use two distinct objects in case a possibly improved sort
     * implementation looks at the address of the objects. */
    PyList_SET_ITEM(list, 0, PyInt_FromLong(0));
    PyList_SET_ITEM(list, 1, PyInt_FromLong(1));

    /* Create the compare function and bind it to our list, so that
     * the list object will be the first parameter to dummy_cmpfunc(). */
    cmpfunc = PyCFunction_New(&dummy_cmpfunc_md, list);
    if (cmpfunc == NULL)
        goto error;

    /* Get the original list.sort() method. */
    sort_meth = PyObject_GetAttrString(list, "sort");
    if (sort_meth == NULL || !PyCFunction_Check(sort_meth))
        goto error;

    /* Sort the list - this will call dummy_cmpfunc() at least once. */
    sort_args = Py_BuildValue("(O)", cmpfunc);
    if (sort_args == NULL)
        goto error;
    sort_res = PyEval_CallObject(sort_meth, sort_args);
    if (sort_res == NULL)
        goto error;

    /* Save sort_meth. */
    if (stablesort_immutable_list_type != NULL)
    {
        PyCFunctionObject *m = (PyCFunctionObject *) sort_meth;
        if (m->m_ml && m->m_ml->ml_meth)
        {
            /* success */
            list_sort_md = m->m_ml;
            sort_meth = NULL;           /* permanently keep a reference */
            result = 0;
        }
    }

error:
    Py_XDECREF(sort_res);
    Py_XDECREF(sort_args);
    Py_XDECREF(sort_meth);
    Py_XDECREF(cmpfunc);
    Py_XDECREF(list);
    return result;
}


/***********************************************************************
// main
************************************************************************/

static PyMethodDef methods[] =
{
    {"sort",       (PyCFunction)do_sort,      METH_VARARGS, SORT_DOC},
    {"install",    (PyCFunction)do_install,   METH_VARARGS, INSTALL_DOC},
    {"uninstall",  (PyCFunction)do_uninstall, METH_VARARGS, UNINSTALL_DOC},
    {NULL, NULL, 0, NULL}
};


static /* const */ char module_documentation[] =
    SORT_DOC "\n"
    INSTALL_DOC "\n"
    UNINSTALL_DOC "\n"
;


#ifdef _MSC_VER
_declspec(dllexport)
#endif
void initstablesort(void)
{
    PyObject *m, *d, *v;

    if (init_internals() != 0) {
        /* this really should not happen... */
        Py_FatalError("stablesort: init_internals()");
        return;
    }

    m = Py_InitModule4("stablesort", methods, module_documentation,
                       NULL, PYTHON_API_VERSION);
    d = PyModule_GetDict(m);

    v = PyString_FromString("Markus F.X.J. Oberhumer (wrapper), Tim Peters (actual sort implementation)");
    PyDict_SetItemString(d, "__author__", v);
    Py_DECREF(v);
    v = PyString_FromString(MODULE_VERSION " (listobject.c CVS 2.134)");
    PyDict_SetItemString(d, "__version__", v);
    Py_DECREF(v);
    v = PyString_FromString(__DATE__ " " __TIME__);
    PyDict_SetItemString(d, "__compile_date__", v);
    Py_DECREF(v);
}


/*
vi:ts=4:et
*/
